/*    Copyright (C) 1999,  Daniel Schulz	   
 * 
 *    This file is part of the TDI library. The TDI library implements  
 *    the generic part of the Thread Debug Interface for POSIX Threads
 *    Implementations - TDI.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Library General Public
 *    License as published by the Free Software Foundation; either
 *    version 2 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this library; if not, write to the Free
 *    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    Report problems and direct all questions to:
 *
 *    tdi-bugs@informatik.hu-berlin.de
 *	
 * $Author: dschulz $ $Date: 1999/05/02 21:31:29 $ $Name: TDI-alpha-release-2 $ File: ./Comm/tdi-comm-sm.c
 *
 */
#include "comm-config.h"
#include "tdi-comm.h"
#include "tdi-comm-sm.h"

#ifdef TDI_SERVER
#include "smalloc.h"     /* secure malloc,calloc,realloc,free,strdup */
#include "tdi-dlgen.h"   /* secure standard functions */
#endif

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>


#define V(x) 


static IpcChannelT SMIpcChannelTemplate = {

  0,          /* id                    */
  (void *) 0, /* channel specific data */
  0,          /* init */ 
  sm_open,
  sm_close,
  sm_send,
  sm_receive
};


int
newSMChannel(IpcChannelT* IpcChan, char* spec) {
  
  /* assign the template */
  *IpcChan = SMIpcChannelTemplate;

  /* init the template */
  if (IpcChan->init)
    return IpcChan->init(IpcChan);

  return IPC_OK;
} 

/*------------------------------------ *
 * sm_open: opens the shared memory    *
 *          and returns the sm_id      *
 *------------------------------------ */ 

/* an open should be possible without a close  *
 * -> therefore no dynamic memory acces should *
 * be happen in open                           */

char      u_data [sizeof(sm_u_data)];
char null_u_data [sizeof(sm_u_data)];

int 
sm_open (IpcChannelT* Chan, void *keyPtr) {

  int sm_id=0;                                /* SM id      */
  long key = (keyPtr)?*((long*)keyPtr):0L;    /* key for SM */

  *((sm_u_data*)u_data) = *((sm_u_data*)null_u_data);
  
  /* FIXME: - there is only one TDI-Channel used now *
   *        - for further channels u_data shold be   *
   *          extended by one dimension              */ 
  Chan->IpcData = u_data;

  ((sm_u_data*)Chan->IpcData)->ReadyToWrite         = SemOpen(key,SEM_FULL);
  ((sm_u_data*)Chan->IpcData)->NewData              = SemOpen(key+1,SEM_EMPTY);
  ((sm_u_data*)Chan->IpcData)->synchronized         = 1;
  ((sm_u_data*)Chan->IpcData)->SMPageSize           = SM_PAGE_SIZE;

  


  /* if there is already a shared memory segment, this shmget call *
   * returns a positive value                                      */ 
  
  
  if ((sm_id=shmget(key,
		    sizeof(char)*((sm_u_data*)Chan->IpcData)->SMPageSize,
		    0))>0) 
    {
      /* we are the first opener */

      return (Chan->ID = sm_id);
    }

  if ((sm_id=shmget(key,
		    sizeof(char)*((sm_u_data*)Chan->IpcData)->SMPageSize,
		    IPC_CREAT|SHM_PERMS))<0) {
    return IPC_UNKNOWN_ID;
  }
  else {
    return (Chan->ID = sm_id);
  }
}

void 
sm_close(IpcChannelT* Chan) {
  shmctl(Chan->ID,IPC_RMID,(struct shmid_ds*)0);
  SemClose(((sm_u_data*)Chan->IpcData)->ReadyToWrite);
  SemClose(((sm_u_data*)Chan->IpcData)->NewData);
  Chan->IpcData=0;
}

/* ------------------------------------------- *
 * the SM header is coded as follows:          *
 * - the first SM_HEADER_BYTES bytes are       *
 *   the length in bytes of the following data *
 * - the next byte is a flag [0 = last page]   *
 * ------------------------------------------- */
     


int 
sm_send(IpcChannelT* Chan, char* data, int length) {

  int i=0;
  char *SMPage=NULL;
  sm_page_header_t header;
  int SMDataSize   = 
    ((sm_u_data*)Chan->IpcData)->SMPageSize - SM_HEADER_BYTES;

  V(printf("\nsm_send:\n"));

  for (i=0; i<length ; i+=SMDataSize) {

    /* how many have to be sent this time */
    memset(&header,0,sizeof(sm_page_header_t));

    header.shore = getpid();

    header.length = ((length - i) >= SMDataSize)? 
      SMDataSize:
    (length - i);
    
    /* is there a page after this */
    if ((length - i) > SMDataSize) {
      header.follows = 1;
    }

    if (!header.length)
      continue;

    if (((sm_u_data*)Chan->IpcData)->synchronized)
      SemGet(((sm_u_data*)Chan->IpcData)->ReadyToWrite);
    
    /* attach memory */
    if ((SMPage=(char*)shmat(Chan->ID,(char*)0,0))<0) {
      return IPC_ERROR;
    };
    
    /* printf("(putmsg) Writing %d slots.\n", SMDataSize); */
    
    /* code the header */
    *((sm_page_header_t*)SMPage) = header; 

    /* write, write and ... */
    memcpy(&SMPage[SM_HEADER_BYTES],&data[i],sizeof(char)*header.length);  


    V(printf("\tsend: %d bytes\n",header.length));

    /* detach memory */
    if (shmdt((char*)SMPage)<0) { 
      return IPC_ERROR;
    }
    
    /* Unlock */
    if (((sm_u_data*)Chan->IpcData)->synchronized)
      SemGive(((sm_u_data*)Chan->IpcData)->NewData);
  }  

  return IPC_OK;
}

int 
sm_receive(IpcChannelT* Chan, char **data, int* received) {
  int i=0,attempt=0;
  unsigned char PageFollows = 1;
  char* SMPage = NULL;
  sm_page_header_t header;
  int SMDataSize   = ((sm_u_data*)Chan->IpcData)->SMPageSize - SM_HEADER_BYTES;
  
  /* ------------------------------------------------------ */
  
  *received = 0; /* prevent chaotic behaviour */
  
  
  V(printf("\nsm_receive:\n"));
  
  /* attach memory */
  if ((SMPage=(char*)shmat(Chan->ID,(char*)0,0))<0) {
    return IPC_ERROR;
  };
  
  for (i=0; PageFollows ; i+=SMDataSize) {
    
    /* it is a ping pong protocol:                    *
     *  - if nothing was read by the other shore, we  *
     *    release the lock                            */
    
    while(attempt++ < TDI_SM_MAX_ATTEMPT_TO_RECEIVE_DATA) {
      
      /* lock the memory segment  */
      if (((sm_u_data*)Chan->IpcData)->synchronized)
	SemGet(((sm_u_data*)Chan->IpcData)->NewData);
      
      /* extract the header */
      header = *((sm_page_header_t*)SMPage);
      
      if (header.shore == getpid()) {
	/* this data packet was written by us *
	 * -> release the memory segment      */
	SemGive(((sm_u_data*)Chan->IpcData)->NewData);
      }
      else 
	break;
    }
    
    if (attempt >= TDI_SM_MAX_ATTEMPT_TO_RECEIVE_DATA) {
      
      /* detach memory */
      if (shmdt((char*)SMPage)<0) { 
	return IPC_ERROR;
      }
      
      return IPC_ERROR;
    }

    /* are there packages left to read */
    PageFollows = header.follows;
    
    /* the sum of all frames for this response */
    (*received)+= header.length;
    
    /* add the frame to the whole response data block */
    memcpy(&((*data)[i]),&SMPage[SM_HEADER_BYTES],header.length);  
    
    V(printf("\treceive: %d bytes\n",header.length));
    
    /* Unlock */
    if (((sm_u_data*)Chan->IpcData)->synchronized)
      SemGive(((sm_u_data*)Chan->IpcData)->ReadyToWrite);
  }
  
  /* detach memory */
  if (shmdt((char*)SMPage)<0) { 
    return IPC_ERROR;
  }

  return IPC_OK;
  
}

/* synchronisaton stuff */

/* semaphore stuff */

/* SemGet */
static struct sembuf P[2] = {
  {0,0,0}, /* wait for sem to become 0 */
  {0,1,0}  /* then increment sem by 1 */
};

/* SemGive */
static struct sembuf V[1] = {
  { 0,-1,IPC_NOWAIT }  /* free the sem */
  /* 0,-1,0 */ 
};

/* HAVE_UNION_SEMUN is part of the configure stuff */ 
#if !HAVE_UNION_SEMUN 
typedef union semun {  
  int val;
  struct semid_ds *buf;
  ushort *array;	
} SemCtlArgT;
#endif

int SemOpen(int key, int option) {
  int semid;
#if !HAVE_UNION_SEMUN
  SemCtlArgT arg;
#else
  union semun arg;
#endif 

  arg.val = (option == SEM_FULL)?1:0;

  /* alternative 2: sem exists */
  if ((semid = semget(key,SEM_CNT,SEM_PERMS))>0) {
    V(printf("sem: second\n"));
    return semid;
  }
  
  /* alternative 1: there is no sem with that key */
  if ((semid = semget(key,SEM_CNT,IPC_CREAT|SEM_PERMS))>0) {
    semctl(semid,SEM_CNT,SETVAL, arg);
    V(printf("sem: first\n"));
    return semid;
  }
  
  /* alternative 3: sth. is really wrong */
  return IPC_ERROR;
}


void SemClose(int semid) {
  semctl(semid,SEM_CNT,IPC_RMID,0);
}

int SemGive(int semid) {
  if (semop(semid,V,1)<0){ 
    return IPC_ERROR;
  }
  return IPC_OK;
};

int SemGet(int semid) {
  if (semop(semid,P,2)<0) { 
    return IPC_ERROR;
  }
  return IPC_OK;
};

