/*    Copyright (C) 1999,  Daniel Schulz	   
 * 
 *    This file is part of the TDI library. The TDI library implements  
 *    the generic part of the Thread Debug Interface for POSIX Threads
 *    Implementations - TDI.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Library General Public
 *    License as published by the Free Software Foundation; either
 *    version 2 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this library; if not, write to the Free
 *    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    Report problems and direct all questions to:
 *
 *    tdi-bugs@informatik.hu-berlin.de
 *	
 * $Author: dschulz $ $Date: 1999/05/03 23:04:08 $ $Name: TDI-alpha-release-2 $ File: smalloc.c
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "smalloc.h"
#include "tdi-dlgen.h"

#define P(x)


#define FREE    0x0
#define ALLOCED 0x1

#define CORE_ADDR void* 
#define CORE_VAL  char

CORE_VAL  smem[SMEM_SIZE];
CORE_ADDR smem_start = &smem[0];

int smem_free = SMEM_SIZE;
int smem_piece_free = SMEM_ENTRYNR-1; /* the first piece is alloced */

/* a piece of registered memory */

struct smem_piece {
  struct smem_piece* prev; /* addr of next smem_piece structure */
  struct smem_piece* next; /* addr of prev smem_piece structure */ 
  CORE_ADDR end;           /* end of this piece */
  int size;                /* the size */
  char flag;               /* ALLOCED ? */
};

/* the global table of aloced and free memory patches */

struct smem_piece smem_list[SMEM_ENTRYNR] = {{(struct smem_piece*)-1,
					      NULL,
					      (CORE_ADDR)-1,
					      0,
					      ALLOCED}};

/* iteration macro */

#define ALL_PIECES(piece) for (piece=&smem_list[0];piece;piece=piece->next)

/* an initializer */
struct smem_piece null_piece; /* NULL piece is used for cleaning */ 

/* at the beginning the first is the last element */
struct smem_piece* last = &smem_list[0];

/* it is not initialized yet */
int initialized = 0;

/* forward declaration */
struct smem_piece* coalesce(struct smem_piece* this, struct smem_piece* next); 

/* if there is a free table entry, return it */
struct smem_piece*
newmempiece() {
  int i;
  for (i=0; i<SMEM_ENTRYNR; i++) {
    if (smem_list[i].prev == NULL) {/* NOTE: first->prev = -1 */
      smem_piece_free --;
      return &smem_list[i];
    }
  }
  return NULL;
}

/* gets the table entry for a specified memory address */
struct smem_piece*
getmempiece(void* ptr) {

  struct smem_piece* piece;
  
  ALL_PIECES(piece) {
    if (piece->end == ptr-1 && piece->next)
      return piece->next;
  }
  return NULL;
}

/* the termination element ends at -1, so that the *
 * first element starts at 0                       */
void 
init() {
  smem_list[0].end = smem_start-1;
  initialized = 1;
}

void*
s_calloc(size_t nmemb, size_t size) {
  void* ret = NULL;
  int i;

  if (!(ret = s_malloc(nmemb*size)))
    return NULL;

  for (i=0;i<nmemb*size;i++) {
    ((char*)ret)[i]=0;
  } 

  return ret;
}

#ifdef __linux__
#define SMALLOC_ALIGNSIZE 16 /* on an x86 it seems to be 1 */
#else
#define SMALLOC_ALIGNSIZE sizeof(void*)
#endif

void*
s_malloc(size_t size) {
  struct smem_piece* piece;
  struct smem_piece* new_piece;
  int aligned_size;

  /* that was a joke */
  if (!size)
    return NULL;

  /* compute an aligned size */
 
  if (size % SMALLOC_ALIGNSIZE)
    {
      /* not aligned */
      aligned_size = ((size/SMALLOC_ALIGNSIZE)+1)*SMALLOC_ALIGNSIZE;
    }
  else
    aligned_size=size;

  /* memory overflow */
  if (smem_free - aligned_size < 0) {
    return NULL;
  }

  if (!initialized)
    init();

  /* look for a free rest */
  
  ALL_PIECES(piece) {
    if (piece->flag==FREE &&     /* a free piece */ 
	piece->size >= aligned_size)     /* greater or equal than size */
	
      {
	/* case A: memory segment is not 1.1 as big */
	if ((float)piece->size <= (float)aligned_size * 1.1)
	  {
	    piece->flag = ALLOCED;
	    smem_free -= piece->size;
	    P(printf(" [c:%.1f s:%.1f m:%.1f ]+ ",(float)piece->size,(float)aligned_size ,(float)aligned_size* 1.1));
	    return piece->prev->end+1;
	  }

	else /* desired size is much less than current size */ 
	  { 
	    /* split the segment */
	    if ((new_piece = newmempiece())!=NULL) {
	      struct smem_piece *a,*b,*c;

	      a=piece;
	      b=new_piece;
	      c=piece->next;
	      
	      b->prev = a;
	      b->next = c;
	      b->size = a->size - aligned_size;
	      b->end  = a->end;

	      a->next = b;
	      a->size = aligned_size;
	      a->end  = a->prev->end + aligned_size;
	      a->flag = ALLOCED;

	      c->prev = b;
	      
	      smem_free -= aligned_size;
	      return a->prev->end+1;
	    }
	    else
	      return NULL;
	  }

      }
  }

  /* trying to get a new mem_piece */
  if ((new_piece = newmempiece())!=NULL) 
    {
      if (last->end + aligned_size > (smem_start+SMEM_SIZE)) 
	{
	  P(printf("0x%x -> address out of bounds.\n",last->end + aligned_size));
	  return NULL;
	}

      /* decrease the global free size */
      smem_free -= aligned_size;
      /* assign the size */
      new_piece->size = aligned_size;
      /* compute the end */
      new_piece->end  = last->end + aligned_size;
      /* mark it as alloced */
      new_piece->flag = ALLOCED;
      /* link it with the last */
      last->next      = new_piece;
      new_piece->prev = last;

      /* we become the last */
      last = new_piece;

      return new_piece->prev->end+1;
    }


  return NULL;
}


void *
s_realloc (void *ptr, size_t size) {
  struct smem_piece* piece = NULL;
  void* new_ptr;

  /* ptr == NULL -> malloc */
  if (ptr == NULL)
    return s_malloc(size);

  /* size == 0 -> free */
  if (size==0) 
    {
      s_free(ptr);
      return NULL; 
    }
  
  /* get the memory peace */
  piece = getmempiece(ptr);

  if (piece->flag == FREE)
    return NULL;

#if 0 /* FIXME: this should be implemented */
  /* if size is <= current size -> shrink the piece */
  

  /* if size > current size -> 
   * a) the next piece in the chain is free and large enough
   *    -> melt them together
   * b) the next piece in the chain is free and not large
   *    enough 
   *    -> malloc a new piece 
   *    && -> copy the contents
   *       -> free the old one
   */
#endif

  new_ptr = s_malloc(size);

  P(printf("old: %x new: %x\n",ptr,new_ptr));

  if (new_ptr) 
    {
      int i;
      /* copy the content */
      for (i=0; i<size;i++) 
	{
	  ((CORE_VAL*)new_ptr)[i] = ((CORE_VAL*)ptr)[i];
	} 
      /* free the old pointer */
      s_free(ptr);
      /* return the new one */
      return new_ptr;
    }
  else 
    /* the malloc failed -> we fail too */
    return NULL;
}

struct smem_piece*
coalesce(struct smem_piece* this, struct smem_piece* next) 
{

  /* pointer swizzling */
  this->next = next->next;
  if (next->next) /* the next element is not the last */
    next->next->prev = this;
  else            /* the next was the last -> we become the last */
    last = this;

  /* end value computing */
  this->end = next->end;

  /* size value computing */
  this->size+= next->size;

  /* compute number of free pieces */
  smem_piece_free++;
  
  /* next is deleted anyway */
  *next = null_piece;

  return this;

}

/* deletes the last piece of memory in the *
 * chain                                   */ 

void 
kick_last() {
  struct smem_piece* piece = last;
  /* the prev becomes the last */
  last = piece->prev;
  /* the next of the prev no longer exists */
  piece->prev->next = NULL;
  /* compute number of free pieces */
  smem_piece_free++;
  /* we remove  us */
  *piece = null_piece;
}

void 
s_free(void *ptr) {
  struct smem_piece* piece = NULL;
  char changed;

  /* search the mem_piece */
  piece = getmempiece(ptr);

  if (!piece) {
    abort();
  }

  piece->flag = FREE;
  smem_free+=piece->size;

  changed = 1;
  while (changed) {
    changed = 0;
    
    /* if the next is FREE too, we fuse them */
    if (piece->prev->flag==FREE) {
      piece = coalesce(piece->prev,piece);
      changed = 1;
    }
    
    /* if the prev is FREE too, we fuse them */
    if (piece->next)
      if (piece->next->flag==FREE) {
	piece = coalesce(piece,piece->next);
	changed = 1;
      }
    
    /* if we were the last, kick us */
    if (last == piece) {
      kick_last();
      return;
    }
  }
}

/* a special case of malloc -> strdup */
char *s_strdup(const char *s) {
  char* new=NULL;
  new = s_malloc(strlen(s)+1);
  sprintf(new,s);
  return new;
}

#define TEST_STRINGS (SMEM_ENTRYNR-20)
#define TEST_STRING_SIZE ((SMEM_SIZE/(SMEM_ENTRYNR+20)))

void s_test_suite() {

  int i,size = 1;
  char* str[TEST_STRINGS];

  if ( TEST_STRING_SIZE < 16) {
    printf ("test string size %d is to small.\n",TEST_STRING_SIZE);
    return;
  }

  for (i=0;i<TEST_STRINGS;i++)
    {
      switch (i%5) 
	{
	case 0: size =  TEST_STRING_SIZE - 5;  break;
	case 1: size =  TEST_STRING_SIZE - 10; break;
	case 2: size =  TEST_STRING_SIZE - 15; break;
	case 3: size =  TEST_STRING_SIZE - 10; break;
	case 4: size =  TEST_STRING_SIZE - 5;  break;
	}
	  
      str[i] = (char*)s_malloc(size);
      {
	int j;
	for (j=0;j<size;j++) {
	  str[i][j] = 65 + (j%26);
	}
	str[i][size-1]=0;
      }
      printf("%s\n",str[i]);
      printf("free: %d\n",smem_free);
    }

  for (i=0;i<TEST_STRINGS;i++)
    {
      switch (i%5) 
	{
	case 2: size =  TEST_STRING_SIZE - 5;  break;
	case 0: size =  TEST_STRING_SIZE - 10; break;
	case 1: size =  TEST_STRING_SIZE - 15; break;
	case 3: size =  TEST_STRING_SIZE - 10; break;
	case 4: size =  TEST_STRING_SIZE - 5;  break;
	}

      printf("old length: %d ",strlen(str[i]));

      str[i] = (char*)s_realloc(str[i],size);
      {
	str[i][size-1]=0;
      }
      printf(" new length: %d",strlen(str[i]));
      printf(" -> smem_free: %d smem_piece_free: %d\n",smem_free,smem_piece_free);
    } 

  for (i=0;i<TEST_STRINGS;i++) {
    printf("free: %d -> free 0x%x ->",smem_free, str[i]);
    s_free(str[i]);
    printf("free: %d\n",smem_free);
  }

}











