/*    Copyright (C) 1999,  Daniel Schulz	   
 * 
 *    This file is part of the TDI library. The TDI library implements  
 *    the generic part of the Thread Debug Interface for POSIX Threads
 *    Implementations - TDI.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Library General Public
 *    License as published by the Free Software Foundation; either
 *    version 2 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this library; if not, write to the Free
 *    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    Report problems and direct all questions to:
 *
 *    tdi-bugs@informatik.hu-berlin.de
 *	
 * $Author: dschulz $ $Date: 1999/05/02 21:31:30 $ $Name: TDI-alpha-release-2 $ File: ./Server/tdi-server.c
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>  

#include "tdi.h"
#include "tdi-server.h"
#include "tdi-comm.h"
#include "tdi-comm-sm.h"
#include "tdi-access.h"

/* this is done to use sigaction_t instead of *
 * struct sigaction to avoid redefinition     * 
 * problems with the sigaction call           *
 * --> see smalloc.h for all redefinitions    */

typedef struct sigaction sigaction_t;

#include "tdi-dlgen.h"

#include "smalloc.h"


#define P(x) 

/* ------------------------------------------------------- *
 * create a static empty PthreadImplEntityAccess structure *
 * to write to                                             *
 * ------------------------------------------------------- */

PthreadImplAttrAccessT      ThreadAttributeAccessEmpty[TA_COUNT];
PthreadImplAttrAccessT      MutexAttributeAccessEmpty[MA_COUNT];
PthreadImplAttrAccessT      CondAttributeAccessEmpty[CA_COUNT];

PthreadImplEntityAccessT PthreadImplAccessEmpty[ET_COUNT] = {
  {
    /* iterators */
    {NULL,NULL},
    /* attr access */
    ThreadAttributeAccessEmpty
  },
  {
    /* iterators */
    {NULL,NULL},
    /* attr access */
    MutexAttributeAccessEmpty
  },
  {
    /* iterators */
    {NULL,NULL},
    /* attr access */
    CondAttributeAccessEmpty
  }
};

/* ENTITY_ACCESS_PARTS is the number of defined function vectors  *
 * in PthreadImplAccessEmpty. It is defined, because the compiler *
 * only recognizes element number exceeding                       */
#define ENTITY_ACCESS_PARTS 3

/* if ENTITY_ACCESS_PARTS is not equal to ET_COUNT (the number of *
 * known entity types) we have a violation                        */
#define ELEMENT_COUNT_VIOLATION (ET_COUNT != ENTITY_ACCESS_PARTS)


/* the ImplEntityAccess structure is filled with Zeros */
PthreadImplEntityAccessT* PthreadImplAccess = PthreadImplAccessEmpty;
int PthreadImplAccessAssigned = 0; /* per process toggle */


/* --------------------------------------------------------------- *
 *         registering attribute access functions                  *
 * --------------------------------------------------------------- */

int tdi_set_iterator_functions (RelTypeT RelType,
				ObjRefT (*getFirst) (),
				ObjRefT (*getNext)  (ObjRefT));

int tdi_set_attr_function      (RelTypeT RelType, 
				AttrTypeT AttributeType, 
				AttrDomainT (*getfunc) (ObjRefT),
				AttrDomainT (*setfunc) (ObjRefT,AttrDomainT)
				); 

/* --------------------------------------------------------------- *
 *         service functions to provide a persistent-ID support    *
 * --------------------------------------------------------------- */

int registerObject               (RelTypeT,ObjRefT);
int unregisterObject             (RelTypeT,ObjRefT);
int isRegistered                 (RelTypeT,ObjRefT);
ObjRefT getFirstObject         (RelTypeT);
ObjRefT getNextObject  (RelTypeT,ObjRefT);

/* -------------------------------------------------------------- *
 * All functions are bundled in a sructure and passed to the      *
 * Pthreads-Impl. Another way is to declare the functions extern  *
 * and let the linker do the job. This is not safe if this TDI    *
 * module is not loaded.                                          *
 * -------------------------------------------------------------- */


/* this is used by the thread implementation to get LUN support *
 * and a global Mutex List                                      */
TDIAuxFuncsT TDIAuxFuncs = { 
  /* stuff to let the pthreads registering its *
   * attribute functions and iterators         */
  tdi_set_attr_function,
  tdi_set_iterator_functions,
  /* object registering service */
  registerObject, 
  unregisterObject, 
  isRegistered,
  getFirstObject,
  getNextObject
};

/* ---------------------------------------------------------------------- */


/* this dinction is called to register the TED functions */
#ifdef TRASH
extern void pthread_TDI_register(TDIAuxFuncsT*);
#else
void (*pthread_TDI_register_func)(TDIAuxFuncsT*) = NULL;
#endif

/* the function that handles catched signals */
void tdi_signal_catcher(int);
/* the function that install the catching function */
void tdi_install_catcher();
/* the function that uninstalls the catching func  */
void tdi_uninstall_catcher();

int TdiServiceRequestHandler(char *Request, char** Response);


int tdi_server(int ID, int IpcType) {
  int length=0,pid=0;
  int RespLength=0;
  char *Request=NULL, *Response=NULL;
  CommChannelT MyChannel;

  if (!tdi_islibopen())
    tdi_libopen();

  /* error checking is done after signal masking to shorten *
   * the critical section between server call and masking   */

  if (ELEMENT_COUNT_VIOLATION /* see above */)
    return -1;

#if 0
  fprintf(stderr,"no selective wait\n");
#endif
  wait(0); /* perhaps there are zombies from the last call: *
	      * FIXME: we have to wait only for the forked    *
	      * comm-task, so waitpid is the better way       */

  if (!PthreadImplAccessAssigned) {
    /* get the supported functions */
#ifdef TRASH
    pthread_TDI_register(&TDIAuxFuncs);
#else
    pthread_TDI_register_func(&TDIAuxFuncs);
#endif
    PthreadImplAccessAssigned = 1;

    if ((ID == -1) && (IpcType == -2)) {
      return 0;
    }
  }

  newTDIChannel(&MyChannel,IpcType);
  
  /* ------------------------------------------- *
   * receiving the Gdb request                   *
   * ------------------------------------------- */
  
  MyChannel.open(&MyChannel,&ID );
  MyChannel.receive(&MyChannel, &Request, &length);
  
  /* do the apropriate things and return the number of *
   * generated messages                                */

  RespLength = TdiAccessRequestHandler(Request,&Response);

  if (RespLength > TDI_COMM_MAX_PACKET_SIZE) { /* we have to fork */
    if ((pid=fork())<0) {
      printf("fork error ");
      return (-1);
    }
    
    if (pid > 0) {
      
      if (Request)
	free(Request);
      
      if (Response) 
	free(Response);
      
      return (0);
    }
    
    /* Child */
    MyChannel.send(&MyChannel,Response,RespLength);
    
    exit(0);
  }

  /* no fork -> easy to handle */
  
  if (Request)
    free(Request);
  
  if (Response) 
    free(Response);
  
  MyChannel.send(&MyChannel,Response,RespLength);

  return (0);

}

/* ------------------------------------ *
 * longterm unique numbering of threads *
 * ------------------------------------ */

#define MAX_THREAD_COUNT 2500
#define MAX_MUTEX_COUNT MAX_THREAD_COUNT 
#define MAX_COND_COUNT   MAX_THREAD_COUNT


int registerThread (ObjRefT);
int registerMutex  (ObjRefT);
int registerCond   (ObjRefT);

int registerObject (RelTypeT RelType, ObjRefT ObjRef) {

  if (!ObjRef || (RelType==ET_UNDEF))
    return 0; /* no ID */
  
  switch(RelType) {
  case ET_THREAD:
    return registerThread(ObjRef);
  case ET_MUTEX:
    return registerMutex(ObjRef);
  case ET_COND:
    return registerCond(ObjRef);
  default:
    fprintf(stderr,"(TDI) RelType %d not supported",RelType); 
    return 0;
  }
}

int unregisterThread (ObjRefT);
int unregisterMutex  (ObjRefT);
int unregisterCond   (ObjRefT);

int 
unregisterObject (RelTypeT RelType, ObjRefT ObjRef) {

  if (!ObjRef || (RelType==ET_UNDEF))
    return 0; /* no ID */
  
  switch(RelType) {
  case ET_THREAD:
    return unregisterThread(ObjRef);
  case ET_MUTEX:
    return unregisterMutex(ObjRef);
  case ET_COND:
    return unregisterCond(ObjRef);
  default:
    fprintf(stderr,"(TDI) RelType %d not supported",RelType); 
    return 0;
  }
}

int isregisteredThread (ObjRefT);
int isregisteredMutex  (ObjRefT);
int isregisteredCond   (ObjRefT);
  
int  
isRegistered (RelTypeT RelType, ObjRefT ObjRef) {

  if (!ObjRef || (RelType==ET_UNDEF))
    return 0; /* no ID */
  
  switch(RelType) {
  case ET_THREAD:
    return isregisteredThread(ObjRef);
  case ET_MUTEX:
    return isregisteredMutex(ObjRef);
  case ET_COND:
    return isregisteredCond(ObjRef);
  default:
    fprintf(stderr,"(TDI) RelType %d not supported",RelType); 
    return 0;
  }
}

ObjRefT getFirstRegisteredThread ();
ObjRefT getFirstRegisteredMutex  ();
ObjRefT getFirstRegisteredCond   ();
  
ObjRefT 
getFirstObject (RelTypeT RelType) {

  if (RelType==ET_UNDEF)
    return 0; /* no ID */
  
  switch(RelType) {
  case ET_THREAD:
    return getFirstRegisteredThread();
  case ET_MUTEX:
    return getFirstRegisteredMutex();
  case ET_COND:
    return getFirstRegisteredCond();
  default:
    fprintf(stderr,"(TDI) RelType %d not supported",RelType); 
    return 0;
  }
}

ObjRefT getNextRegisteredThread (ObjRefT);
ObjRefT getNextRegisteredMutex  (ObjRefT);
ObjRefT getNextRegisteredCond   (ObjRefT);
  
ObjRefT 
getNextObject (RelTypeT RelType, ObjRefT Last) {
  if (!Last || (RelType==ET_UNDEF))
    return 0; /* no ID */
  
  switch(RelType) {
  case ET_THREAD:
    return getNextRegisteredThread (Last);
  case ET_MUTEX:
    return getNextRegisteredMutex  (Last);
  case ET_COND:
    return getNextRegisteredCond   (Last);
  default:
    fprintf(stderr,"(TDI) RelType %d not supported",RelType); 
    return 0;
  }
}

/* -------------------------------------------------------------- *
 * specific registering functions                                 *
 * -------------------------------------------------------------- */

struct id_info {
  void* mem_addr;
  int   ID;
};

struct id_info TDIThreadTable[MAX_THREAD_COUNT]; 
struct id_info TDIMutexTable [MAX_MUTEX_COUNT];
struct id_info TDICondTable  [MAX_COND_COUNT];

int ThreadSerialNumber = 1; /* no ID 0 */
int MutexSerialNumber  = 1;
int CondSerialNumber   = 1;


int
registerThread(ObjRefT threadPtr) 
{
  int Id=0;
  int tableIndex=0;
  
  Id=isregisteredThread(threadPtr);
  
  if(Id)
    return Id;

  for (tableIndex=0;tableIndex<MAX_THREAD_COUNT;tableIndex++) {
    if (TDIThreadTable[tableIndex].ID == 0) {
      TDIThreadTable[tableIndex].ID = 
	(ThreadSerialNumber+1!=0)?ThreadSerialNumber++:1;
      TDIThreadTable[tableIndex].mem_addr = threadPtr;
      return TDIThreadTable[tableIndex].ID;
    }
  }
  return 0;
}

int
unregisterThread(ObjRefT threadPtr) {
  int tableIndex=0;
  int retId = 0;
  
  for (tableIndex=0;tableIndex<MAX_THREAD_COUNT;tableIndex++) {
    if (TDIThreadTable[tableIndex].mem_addr==threadPtr) {
      TDIThreadTable[tableIndex].mem_addr=NULL;
      retId = TDIThreadTable[tableIndex].ID;
      TDIThreadTable[tableIndex].ID=0;
      return retId;
    }
  }
  return 0;
}

int 
isregisteredThread(ObjRefT threadPtr) {
  int tableIndex=0;

  for (tableIndex=0;tableIndex<MAX_THREAD_COUNT;tableIndex++) {
    if (TDIThreadTable[tableIndex].mem_addr==threadPtr) {
      return TDIThreadTable[tableIndex].ID;
      
    }
  }
  return 0;
}

ObjRefT getFirstRegisteredThread () {
  int tableIndex=0;
  
  for (tableIndex=0;tableIndex<MAX_THREAD_COUNT;tableIndex++) {
    if ((TDIThreadTable[tableIndex].mem_addr!=NULL)&&
	(TDIThreadTable[tableIndex].ID!=0)) {
      return TDIThreadTable[tableIndex].mem_addr;
      
    }
  }
  return NULL;
}

ObjRefT getNextRegisteredThread  (ObjRefT lastThreadPtr) {
  int tableIndex=0;
  int found_last = 0;
  
  for (tableIndex=0;tableIndex<MAX_THREAD_COUNT;tableIndex++) {
    if (TDIThreadTable[tableIndex].mem_addr==lastThreadPtr) {
      found_last = 1;
      continue;
    }
    if ((TDIThreadTable[tableIndex].mem_addr!=NULL) &&
	found_last)
      return TDIThreadTable[tableIndex].mem_addr;
  }
  return NULL;
}






int
registerMutex(ObjRefT mutexPtr) 
{
  int Id=0;
  int tableIndex=0;
  
  Id=isregisteredMutex(mutexPtr);

  if(Id) /* 0 is not a valid Id */
    return Id;

  for (tableIndex=0;tableIndex<MAX_MUTEX_COUNT;tableIndex++) {
    if (TDIMutexTable[tableIndex].ID == 0) {
      TDIMutexTable[tableIndex].ID = 
	(MutexSerialNumber+1!=0)?MutexSerialNumber++:1;
      TDIMutexTable[tableIndex].mem_addr = mutexPtr;
      P(printf("register mutex\n"));
      return TDIMutexTable[tableIndex].ID;
    }
  }
  return 0;
}

int
unregisterMutex(ObjRefT mutexPtr) {
  int tableIndex=0;
  int retId = 0;
  
  for (tableIndex=0;tableIndex<MAX_MUTEX_COUNT;tableIndex++) {
    if (TDIMutexTable[tableIndex].mem_addr==mutexPtr) {
      TDIMutexTable[tableIndex].mem_addr=NULL;
      retId = TDIMutexTable[tableIndex].ID;
      TDIMutexTable[tableIndex].ID=0;
      return retId;
    }
  }
  return 0;
}

int 
isregisteredMutex(ObjRefT mutexPtr) {
  int tableIndex=0;
  
  for (tableIndex=0;tableIndex<MAX_MUTEX_COUNT;tableIndex++) {
    if (TDIMutexTable[tableIndex].mem_addr==mutexPtr) {
      return TDIMutexTable[tableIndex].ID;
      
    }
  }
  return 0;
}

ObjRefT getFirstRegisteredMutex () {
  int tableIndex=0;
  
  for (tableIndex=0;tableIndex<MAX_MUTEX_COUNT;tableIndex++) {
    if ((TDIMutexTable[tableIndex].mem_addr!=NULL)&&
	(TDIMutexTable[tableIndex].ID!=0)) {
      return TDIMutexTable[tableIndex].mem_addr;
      
    }
  }
  return NULL;
}

ObjRefT getNextRegisteredMutex  (ObjRefT lastMutexPtr) {
  int tableIndex=0;
  int found_last = 0;

  for (tableIndex=0;tableIndex<MAX_MUTEX_COUNT;tableIndex++) {
    if (TDIMutexTable[tableIndex].mem_addr==lastMutexPtr) {
      found_last = 1;
      continue;
    }
    if ((TDIMutexTable[tableIndex].mem_addr!=NULL) &&
	found_last)
      return TDIMutexTable[tableIndex].mem_addr;
  }
  return NULL;
}

/* condition variables */

int
registerCond(ObjRefT condPtr) 
{
  int Id=0;
  int tableIndex=0;
  
  Id=isregisteredCond(condPtr);

  if (Id)
    return Id;

  for (tableIndex=0;tableIndex<MAX_COND_COUNT;tableIndex++) {
    if (TDICondTable[tableIndex].ID == 0) {
      TDICondTable[tableIndex].ID = 
	(CondSerialNumber+1!=0)?CondSerialNumber++:1;
      TDICondTable[tableIndex].mem_addr = condPtr;
      P(printf("register cond\n"));
      return TDICondTable[tableIndex].ID;
    }
  }
  return 0;
}

int
unregisterCond(ObjRefT condPtr) {
  int tableIndex=0;
  int retId = 0;
  
  for (tableIndex=0;tableIndex<MAX_COND_COUNT;tableIndex++) {
    if (TDICondTable[tableIndex].mem_addr==condPtr) {
      TDICondTable[tableIndex].mem_addr=NULL;
      retId = TDICondTable[tableIndex].ID;
      TDICondTable[tableIndex].ID=0;
      return retId;
    }
  }
  return 0;
}

int 
isregisteredCond(ObjRefT condPtr) {
  int tableIndex=0;
  
  for (tableIndex=0;tableIndex<MAX_COND_COUNT;tableIndex++) {
    if (TDICondTable[tableIndex].mem_addr==condPtr) {
      return TDICondTable[tableIndex].ID;
      
    }
  }
  return 0;
}

ObjRefT getFirstRegisteredCond () {
  int tableIndex=0;
  
  for (tableIndex=0;tableIndex<MAX_COND_COUNT;tableIndex++) {
    if ((TDICondTable[tableIndex].mem_addr!=NULL)&&
	(TDICondTable[tableIndex].ID!=0)) {
      return TDICondTable[tableIndex].mem_addr;
      
    }
  }
  return NULL;
}

ObjRefT getNextRegisteredCond  (ObjRefT lastCondPtr) {
  int tableIndex=0;
  int found_last = 0;

  for (tableIndex=0;tableIndex<MAX_COND_COUNT;tableIndex++) {
    if (TDICondTable[tableIndex].mem_addr==lastCondPtr) {
      found_last = 1;
      continue;
    }
    if ((TDICondTable[tableIndex].mem_addr!=NULL) &&
	found_last)
      return TDICondTable[tableIndex].mem_addr;
  }
  return NULL;
}


/* ---------------------------------------- *
 * the Pthread Package registers its own    *
 * entitity set iterator functions by       *
 * using tdi_set_iterator_functions         *
 * ---------------------------------------- */

int 

tdi_set_iterator_functions (int EntityType,
			    ObjRefT (*getFirst) (),
			    ObjRefT (*getNext)  (ObjRefT))
{
  if (EntityType<0 ||  EntityType > ET_COUNT)
    return TDI_ACCESS_ERROR;

  if (getFirst) {
    PthreadImplAccess[EntityType].EntityContainerAccess.getFirst = getFirst;
  }
  
  if (getNext) {
    PthreadImplAccess[EntityType].EntityContainerAccess.getNext = getNext;
  }

  return TDI_ACCESS_OK;

}

/* ---------------------------------------- *
 * the Pthread Package registers its own    *
 * entity attribute access functions by     *
 * using tdi_set_attr_function              *
 * ---------------------------------------- */

int 
tdi_set_attr_function(RelTypeT RelType, 
		      AttrTypeT AttributeType, 
		      AttrDomainT (*get_func) (ObjRefT),
		      AttrDomainT (*set_func) (ObjRefT,AttrDomainT)
		      ) 
{
  /* Variables to iterate over attribute types */
  int AttrEnd=0;

  /* error test */
  if (RelType == ET_UNDEF || !(get_func || set_func))
    return TDI_ACCESS_ERROR;
  
  /* set the appropriate end for attr type */ 
  switch (RelType) {
  case ET_THREAD: 
    AttrEnd=TA_COUNT; break;
  case ET_MUTEX:
    AttrEnd=MA_COUNT; break;
  case ET_COND:
    AttrEnd=CA_COUNT; break;
  default:
    return TDI_ACCESS_ERROR;
  }

  if (AttributeType < 0 || AttributeType > AttrEnd) {
    return TDI_ACCESS_ERROR;
  }
  
  if (get_func) { 
    PthreadImplAccess[RelType].EntityAttributeAccess[AttributeType].get =
      get_func;
  }
  
#if 1
  if (set_func) {
    PthreadImplAccess[RelType].EntityAttributeAccess[AttributeType].set =
      set_func;
  }

#endif
  return TDI_ACCESS_OK;
}


